/*  equver - solver of non-linear system of equations given in infix form
    Copyright (C) 2009  Tibor 'Igor2' Palinkas

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

    Project page: http://repo.hu/projects/equver
    Contact:      equver@igor2.repo.hu
*/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <math.h>
#include <minpack.h>
#include "equver.h"
#include "optimize.h"
#include "ipath.h"

static char *vnames[MAXVARS];
static vflags_t vflags[MAXVARS];
static double x[MAXVARS], x_shadow[MAXVARS];
static int vars=0;

static code_t *eq[MAXEQUS];
static char *eq_txt[MAXEQUS];
char useful_eq[MAXEQUS];

static double deviation[MAXEQUS];
static int eqs = 0;
int verbose = 0;

/* these paths are added as if the user specified them with -I, in this order: */
const char *ipath_defaults[] = {
	"/usr/share/equver",
	"~/.equver",
	"./equver",
	".",
	NULL
};

#define SWAP(i1, i2, ty) \
do { \
	ty tmp; \
	tmp = i1; \
	i1 = i2; \
	i2 = tmp; \
} while(0)

void swap_eq(int e1, int e2)
{
	SWAP(eq[e1], eq[e2], code_t *);
	SWAP(eq_txt[e1], eq_txt[e2], char *);
	SWAP(useful_eq[e1], useful_eq[e2], char);
}

void swap_vars(int e1, int e2)
{
	int n;
	code_t *c;
	for(n = 0; n < eqs; n++) {
		for(c = eq[n]; c != NULL; c = c->next) {
			if (c->c == OPC_VAR) {
				if (c->data.var == e1)
					c->data.var = e2;
				else if (c->data.var == e2)
					c->data.var = e1;
			}
		}
	}

	SWAP(vnames[e1],   vnames[e2],   char *);
	SWAP(vflags[e1],   vflags[e2],   vflags_t);
	SWAP(x[e1],        x[e2],        double);
	SWAP(x_shadow[e1], x_shadow[e2], double);
}

int append_alloced = 0, append_used = 0, append_term = 0;
char *append_buff = NULL;

void append(const char *s)
{
	int l;
	char *end;

	if (append_term)
		return;

	end = strchr(s, '\n');
	if (end != NULL) {
		*end = '\0';
		append_term = 1;
	}

	l = strlen(s);
	if (l > 0) {
		if (append_used+l+1 >= append_alloced) {
			append_alloced = append_used+l + 128;
			append_buff = realloc(append_buff, append_alloced);
		}
		memcpy(append_buff + append_used, s, l+1);
		append_used += l;
	}

	if ((append_term) && (append_used == 0))
		append_term = 0;
}

void append_reset()
{
	append_used = 0;
	append_term = 0;
}

extern int yylex_destroy (void);

code_t *alloc_bif(opcode_t code, bif_t *bif, code_t *parm1, code_t *parm2)
{
	code_t *c, *end;

	c = malloc(sizeof(code_t));
	c->c = code;
	c->data.bif = bif;
	c->next = parm1;

	if (parm2 != NULL) {
		for(end = parm1; end->next != NULL; end = end->next) ;
		end->next = parm2;
	}

	return c;
}

code_t *alloc_op(operator_t op, code_t *c1, code_t *c2)
{
	code_t *c, *end;
	c = malloc(sizeof(code_t));
	c->c = OPC_OP;
	c->data.op = op;
	c->next = c1;
	for(end = c1; end->next != NULL; end = end->next) ;
	end->next = c2;
	return c;
}

code_t *alloc_num(double num)
{
	code_t *c;
	c = malloc(sizeof(code_t));
	c->c = OPC_NUM;
	c->data.num = num;
	c->next = NULL;
	return c;
}

static int var_find(const char *vname)
{
	int n;
	for(n = 0; n < vars; n++)
		if (strcmp(vnames[n], vname) == 0)
			return n;
	return -1;
}

int alloc_var_(const char *vname, vflags_t orflags)
{
	int v;
	v = var_find(vname);
	if (v < 0) {
		v = vars;
		vars++;
		vnames[v] = strdup(vname);
	}
	vflags[v] |= orflags;
	return v;
}

code_t *alloc_var(const char *vname, vflags_t orflags)
{
	code_t *c;
	int v;

	v = alloc_var_(vname, orflags);
	c = malloc(sizeof(code_t));
	c->c = OPC_VAR;
	c->data.var = v;
	c->next = NULL;
	return c;
}

static const char *vflag_names[] = {
	"silent",          "omit var from result printout",
	"scale_suffix",    "may be used as scale suffix to scale output variables",
	"constant",        "the ID is a constant, not a variable",
	NULL,       NULL
};

void var_set_flag(const char *vname, const char *flags)
{
	const char **f;
	int b;

	for(f = vflag_names, b = 1; *f != NULL; f+=2,b<<=1) {
		if (strcmp(*f, flags) == 0) {
			alloc_var_(vname, b);
			return;
		}
	}
	fprintf(stderr, "Error: unknown variable flag: '%s'\n", flags);
}

void free_vars()
{
	int n;
	for(n = 0; n < vars; n++)
		free(vnames[n]);
}

void set_initial(const char *vname, double value)
{
	int v;
	v = alloc_var_(vname, VF_NONE);
	x[v] = value;
}

void print_code(code_t *code)
{
	for(; code != NULL; code = code->next) {
		switch(code->c) {
			case OPC_OP:
				printf("OP:  %d\n", code->data.op);
				break;
			case OPC_NUM:
				printf("NUM: %f\n", code->data.num);
				break;
			case OPC_VAR:
				printf("VAR: [%d] %s\n", code->data.var, vnames[code->data.var]);
				break;
			case OPC_BIF:
				printf("BIF: (TODO)\n");
				break;
			case OPC_BIF2:
				printf("BIF2: (TODO)\n");
				break;
		}
	}
}

void fin_eq(code_t *code)
{
	code_t *stack[1024], *c;
	int n;

	if (verbose > 1) {
		printf("EQ %d:\n", eqs);
		print_code(code);
	}

	/* reverse order of code */
	for(n = 0, c = code; c != NULL; n++, c = c->next)
		stack[n] = c;
	n--;
	code = stack[n];
	for(; n > 0; n--)
		stack[n]->next = stack[n-1];

	stack[0]->next = NULL;

	if (verbose > 3) {
		printf("rev:\n");
		print_code(code);
	}

	eq[eqs] = code;

	eq_txt[eqs] = append_buff;
	append_buff = NULL;
	append_alloced = 0;
	append_reset();

	if (verbose > 1)
		printf("\n");
	eqs++;
}

void free_eq(code_t *c)
{
	code_t *next;
	for(; c != NULL; c = next) {
		next = c->next;
		free(c);
	}
}

void yyerror(char *s)
{
	fprintf(stderr, "Parser error %s\n", s);
	abort();
}

int yywrap()
{
	return 1;
}

double eq_eval(code_t *eq, double *v)
{
	int sp, n;
	double stack[1024];
	code_t *c;

	sp = 0;
	for(c = eq; c != NULL; c = c->next) {
		switch(c->c) {
			case OPC_OP:
				assert (sp > 0);
				sp--;

				if (verbose > 2) {
					printf("op %d on %f and %f (sp=%d)\n", c->data.op, stack[sp-1], stack[sp], sp);
					printf(" Stack: ");
					for(n = 0; n < 8; n++)
						printf("%f ", stack[n]);
					printf("\n");
				}

				switch(c->data.op) {
					case OP_ADD: stack[sp-1] += stack[sp]; break;
					case OP_SUB: stack[sp-1] -= stack[sp]; break;
					case OP_MUL: stack[sp-1] *= stack[sp]; break;
					case OP_DIV: stack[sp-1] /= stack[sp]; break;
				}
				break;
			case OPC_NUM:
				stack[sp] = c->data.num;
				sp++;
				break;
			case OPC_VAR:
				stack[sp] = v[c->data.var];
				sp++;
				break;
			case OPC_BIF:
				stack[sp-1] = c->data.bif(stack[sp-1]);
				break;
			case OPC_BIF2:
				stack[sp-2] = c->data.bif2(stack[sp-1], stack[sp-2]);
				sp--;
				break;
		}
	}
	sp--;
	assert(sp == 0);
	return stack[sp];
}

void eval(double *v)
{
	int e;
	double d;

	printf("Eval:\n");
	for(e = 0; e < eqs; e++) {
		d = eq_eval(eq[e], v);
		printf("eq %d -> %f\n", e, d);
	}
}

void fcn(int *funcs, int *vars, double *x, double *fvec, int *iflag)
{
		int f;
		if (verbose > 1)
			printf("fcn\n");
		for(f = 0; f < *funcs; f++) {
			deviation[f] = fvec[f] = eq_eval(eq[f], x);
			if (verbose > 1)
				printf(" %d -> %.20f %s\n", f, fvec[f], eq_txt[f]);
		}
}

void print_var(const char *name, const char *op, double val)
{
	double best;
	int n, bestn = -1;

	if (val >= 1)
		best = val;
	else
		best = 10000000000000000;

	for(n = 0; n < vars; n++) {
		if (vflags[n] & VF_SCALE_SUFFIX) {
			double scaled, sv;
			if (vflags[n] & VF_CONSTANT)
				sv = x_shadow[n];
			else
				sv = x[n];
			scaled = val / sv;
			if ((scaled > 1) && ((scaled < best))) {
				bestn = n;
				best = scaled;
			}
		}
	}
	printf("%s %s %.16f", name, op, val);
	if (bestn != -1)
		printf(" %f %s", best, vnames[bestn]);
	printf("\n");
}

void help(const char *procname)
{
	const char **s;

	printf("%s - non-linear equation solver\n", procname);
	printf("%s reads a system of equations written in infix form on stdin and\n", procname);
	printf("prints a valid root on stdout.\n");
	printf("\n");
	printf("Usage: %s [-v] [-h] [-c] [-i name value] [-I path] [-F vr fl]\n", procname);
	printf(" -v       increase verbosity level (multiple -v arguments can be used)\n");
	printf(" -h       prints this help screen\n");
	printf(" -i       sets the initial value of a variable called 'name' to value\n");
	printf(" -I path  insert path in front of the include search path\n");
	printf(" -I -     reset search paths\n");
	printf(" -F vr fl set flag fl for variable vr\n");
	printf(" -c       do not print constants\n");
	printf("\n");

	printf("Variable flags:\n");
	for(s = vflag_names; *s != NULL; s+=2) {
		printf("  %-16s %s\n", s[0], s[1]);
	}
}

int main(int argc, char *argv[])
{
	int funcs, ret, nonempty;
	int n, maxfev, mode, nprint, info, nfev, ldfjac;
	int *ipvt;
	double d, ftol, xtol, gtol, epsfcn, factor;
	double *fvec, *diag, *fjac, *qtf, *wa1, *wa2, *wa3, *wa4;
	char *cmd, *end;
	int one=1, tries=0, print_consts = 1, num_real_vars;
	const char **ip;

	memset(x, 0, sizeof(x));

	for(ip = ipath_defaults; *ip != NULL; ip++)
		ipath_insert(*ip);

	for(n = 1; n < argc; n++) {
		cmd = argv[n];
		while(*cmd == '-')
			cmd++;

		switch(*cmd) {
			case 'v':
				verbose++;
				break;
			case 'c':
				print_consts = 0;
				break;
			case 'h':
				help(argv[0]);
				return 1;
			case 'F':
				var_set_flag(argv[n+1], argv[n+2]);
				break;
			case 'I':
				if (cmd[1] == '\0') {
					n++;
					cmd = argv[n];
				}
				else
					cmd++;
				if (strcmp(cmd, "-") == 0)
					ipath_reset();
				else
					ipath_insert(cmd);
				break;
			case 'i':
				if ((argv[n+1] == NULL) || (argv[n+2] == NULL)) {
					fprintf(stderr, "Error: -i needs 2 arguments, a variable name and a value\n");
					return 1;
				}
				d = strtod(argv[n+2], &end);
				if (*end != '\0') {
					fprintf(stderr, "Error: invalid floating point number '%s' for -i\n", argv[n+2]);
					return 1;
				}
				if (*argv[n+1] == '$') {
					fprintf(stderr, "Error: can't set special variables ($) with -i\n");
					return 1;
				}
				set_initial(argv[n+1], d);
				n += 2;
				break;
		}
	}

	yyparse(stdin);

	nonempty = optimize(eq, &eqs, x, vnames, &vars, vflags, useful_eq, &num_real_vars);
	memcpy(x_shadow, x, sizeof(x));

	if (nonempty != 0) {
		funcs = nonempty;

		fvec = malloc(sizeof(double) * funcs);
		diag = malloc(sizeof(double) * num_real_vars);
		fjac = malloc(sizeof(double) * num_real_vars * funcs);
		qtf  = malloc(sizeof(double) * num_real_vars);
		wa1  = malloc(sizeof(double) * num_real_vars);
		wa2  = malloc(sizeof(double) * num_real_vars);
		wa3  = malloc(sizeof(double) * num_real_vars);
		wa4  = malloc(sizeof(double) * funcs);
		ipvt = malloc(sizeof(int) * num_real_vars);


  /*      set ftol and xtol to the square root of the machine */
  /*      and gtol to zero. unless high solutions are */
  /*      required, these are the recommended settings. */

		ftol = sqrt(dpmpar_(&one));
		xtol = sqrt(dpmpar_(&one));
		gtol = 0.;

/*	ftol=xtol = 0.0000001;*/

		maxfev = 8000;
		epsfcn = 0.;
		mode = 1;
		factor = 1.e2;
		nprint = 0;
	ldfjac = funcs;
retry:;
		lmdif_(&fcn, &funcs, &num_real_vars, x, fvec, &ftol, &xtol, &gtol, &maxfev, &epsfcn, 
			diag, &mode, &factor, &nprint, &info, &nfev, fjac, &ldfjac, 
			ipvt, qtf, wa1, wa2, wa3, wa4);
	}
	else
		info = 1;

	if (verbose > 1)
		printf("Result: %d (f=%d, v=%d)\n", info, funcs, vars);
	ret = 1;
	switch(info) {
		case 0:
			printf("# ERROR invalid input\n");
			if (funcs < vars)
				printf("# ERROR (not enough equations)\n");
			break;
		case 4:
			if (tries < 32) {
				d = sqrt(xtol);
				for(n = 0; n < vars; n++)
					x[n] += d;
				tries++;
				goto retry;
			}
			printf("# ERROR fvec is orthogonal to the columns of the  Jacobian  to  machine precision.\n");
			break;
		case 5:
			printf("# ERROR number of calls to fcn has reached or exceeded %d*(n+1).\n", maxfev);
			break;
		default:
			{
				int broken = 0, first = 1;
				double maxdev;

				for(n = 0; n < vars; n++) {
					double v;

					if (vflags[n] & VF_CONSTANT)
						continue;

					v = fabs(x[n]);
					if (first || (v < maxdev)) {
						maxdev = v;
						first = 0;
					}
				}
				maxdev /= 1000;

				for(n = 0; n < eqs; n++)
					if (useful_eq[n]) {
						if (fabs(deviation[n]) > maxdev)
							broken++;
						if (verbose > 0)
							printf("Dev % 4d %.16f\t%s\n", n, deviation[n], eq_txt[n]);
					}

				if (verbose > 0)
					printf("\n");

				if (broken)
					printf("#NOTE: deviations are too big (bigger than %.16f), solution is probably not valid!\n", maxdev);
			}
			
			/* print variables; for those variables marked as constant by the optimizer,
			   use shadow value: x[] may have been changed by the solver to figure it
			   didn't matter, still it is not the same value as the constant we
			   should print. */
			for(n = 0; n < vars; n++) {
				if (vflags[n] & VF_SILENT)
					continue;
				if (vflags[n] & VF_CONSTANT) {
					if (print_consts)
						print_var(vnames[n], ".=", x_shadow[n]);
				}
				else
					print_var(vnames[n], "=", x[n]);
			}
			ret = 0;
	}

	free_vars();
	for(n = 0; n < eqs; n++) {
		free_eq(eq[n]);
		free(eq_txt[n]);
	}

	if (nonempty) {
		free(fvec);
		free(diag);
		free(fjac);
		free(qtf);
		free(wa1);
		free(wa2);
		free(wa3);
		free(wa4);
		free(ipvt);
	}

	yylex_destroy();
	return ret;
}
